/* Copyright (C) 2013-2016 Vincent Forest (vaplv@free.fr)
 *
 * The RSys library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * The RSys library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with the RSys library. If not, see <http://www.gnu.org/licenses/>. */

#include "float33.h"
#include "rsys.h"
#include <float.h>

#define CHECK_F3(A, B)                                                         \
  {                                                                            \
    const float* a__ = (A);                                                    \
    const float* b__ = (B);                                                    \
    int i__;                                                                   \
    FOR_EACH(i__, 0, 3)                                                        \
      CHECK(a__[i__], b__[i__]);                                               \
  } (void)0
#define CHECK_F33(A, B)                                                        \
  {                                                                            \
    const float* a__ = (A);                                                    \
    const float* b__ = (B);                                                    \
    int i__;                                                                   \
    FOR_EACH(i__, 0, 9)                                                        \
      CHECK(a__[i__], b__[i__]);                                               \
  } (void)0

int
main(int argc, char** argv)
{
  float a[9], b[9], dst[9], c[9];
  int i;
  (void)argc, (void)argv;

  f33_set(a, f33_splat(c, -1.f));
  FOR_EACH(i, 0, 9) {
    CHECK(a[i], -1.f);
    CHECK(c[i], -1.f);
  }
  f33(a, 0.f, 1.f, 2.f, 3.f, 4.f, 5.f, 6.f, 7.f, 8.f);
  FOR_EACH(i, 0, 9) {
    CHECK(a[i], (float)i);
  }
  CHECK_F33
    (f33_set_identity(a),
     f33(c, 1.f, 0.f, 0.f, 0.f, 1.f, 0.f, 0.f, 0.f, 1.f));
  CHECK_F33(a, f33(c, 1.f, 0.f, 0.f, 0.f, 1.f, 0.f, 0.f, 0.f, 1.f));

  f33_splat(a, -1.f);
  CHECK_F33
    (f33_set_row(a, f3(c, 0.f, 1.f, 2.f), 0),
     f33(c, 0.f, -1.f, -1.f, 1.f, -1.f, -1.f, 2.f, -1.f, -1.f));
  CHECK_F33(a, f33(c, 0.f, -1.f, -1.f, 1.f, -1.f, -1.f, 2.f, -1.f, -1.f));
  CHECK_F33
    (f33_set_row(a, f3(c, 3.f, 4.f, 5.f), 1),
     f33(c, 0.f, 3.f, -1.f, 1.f, 4.f, -1.f, 2.f, 5.f, -1.f));
  CHECK_F33(a, f33(c, 0.f, 3.f, -1.f, 1.f, 4.f, -1.f, 2.f, 5.f, -1.f));
  CHECK_F33
    (f33_set_row(a, f3(c, 6.f, 7.f, 8.f), 2),
     f33(c, 0.f, 3.f, 6.f, 1.f, 4.f, 7.f, 2.f, 5.f, 8.f));
  CHECK_F33(a, f33(c, 0.f, 3.f, 6.f, 1.f, 4.f, 7.f, 2.f, 5.f, 8.f));

  CHECK_F33
    (f33_transpose(a, a),
     f33(c, 0.f, 1.f, 2.f, 3.f, 4.f, 5.f, 6.f, 7.f, 8.f));
  CHECK_F33(a, f33(c, 0.f, 1.f, 2.f, 3.f, 4.f, 5.f, 6.f, 7.f, 8.f));
  CHECK_F33
    (f33_transpose(b, a),
     f33(c, 0.f, 3.f, 6.f, 1.f, 4.f, 7.f, 2.f, 5.f, 8.f));
  CHECK_F33(b, f33(c, 0.f, 3.f, 6.f, 1.f, 4.f, 7.f, 2.f, 5.f, 8.f));

  f33_splat(a, -1.f);
  CHECK_F33
    (f33_set_col(a, f3(c, 0.f, 1.f, 2.f), 0),
     f33(c, 0.f, 1.f, 2.f, -1.f, -1.f, -1.f, -1.f, -1.f, -1.f));
  CHECK_F33(a, f33(c, 0.f, 1.f, 2.f, -1.f, -1.f, -1.f, -1.f, -1.f, -1.f));
  CHECK_F33
    (f33_set_col(a, f3(c, 3.f, 4.f, 5.f), 1),
     f33(c, 0.f, 1.f, 2.f, 3.f, 4.f, 5.f, -1.f, -1.f, -1.f));
  CHECK_F33(a, f33(c, 0.f, 1.f, 2.f, 3.f, 4.f, 5.f, -1.f, -1.f, -1.f));
  CHECK_F33
    (f33_set_col(a, f3(c, 6.f, 7.f, 8.f), 2),
     f33(c, 0.f, 1.f, 2.f, 3.f, 4.f, 5.f, 6.f, 7.f, 8.f));
  CHECK_F33(a, f33(c, 0.f, 1.f, 2.f, 3.f, 4.f, 5.f, 6.f, 7.f, 8.f));

  CHECK_F3(f33_row(b, a, 0), f3(c, 0.f, 3.f, 6.f));
  CHECK_F3(b, f3(c, 0.f, 3.f, 6.f));
  CHECK_F3(f33_row(b, a, 1), f3(c, 1.f, 4.f, 7.f));
  CHECK_F3(b, f3(c, 1.f, 4.f, 7.f));
  CHECK_F3(f33_row(b, a, 2), f3(c, 2.f, 5.f, 8.f));
  CHECK_F3(b, f3(c, 2.f, 5.f, 8.f));

  CHECK_F3(f33_col(b + 2, a, 0), f3(c, 0.f, 1.f, 2.f));
  CHECK_F3(b + 2, f3(c, 0.f, 1.f, 2.f));
  CHECK_F3(f33_col(b + 1, a, 1), f3(c, 3.f, 4.f, 5.f));
  CHECK_F3(b + 1, f3(c, 3.f, 4.f, 5.f));
  CHECK_F3(f33_col(b + 5, a, 2), f3(c, 6.f, 7.f, 8.f));
  CHECK_F3(b + 5, f3(c, 6.f, 7.f, 8.f));

  CHECK_F3(f33_col_ptr(a, 0), f3(c, 0.f, 1.f, 2.f));
  CHECK_F3(f33_col_ptr(a, 1), f3(c, 3.f, 4.f, 5.f));
  CHECK_F3(f33_col_ptr(a, 2), f3(c, 6.f, 7.f, 8.f));
  CHECK_F3(f33_col_cptr(a, 0), f3(c, 0.f, 1.f, 2.f));
  CHECK_F3(f33_col_cptr(a, 1), f3(c, 3.f, 4.f, 5.f));
  CHECK_F3(f33_col_cptr(a, 2), f3(c, 6.f, 7.f, 8.f));

  f33(a, 1.f, 2.f, 3.f, 4.f, 5.f, 6.f, 7.f, 8.f, 9.f);
  CHECK_F3(f33_mulf3(dst, a, f3(c, 1.f, 2.f, 3.f)), f3(c, 30.f, 36.f, 42.f));
  CHECK_F3(dst, f3(c, 30.f, 36.f, 42.f));
  CHECK_F3(f3_mulf33(dst, f3(c, 1.f, 2.f, 3.f), a), f3(c, 14.f, 32.f, 50.f));
  CHECK_F3(dst, f3(c, 14.f, 32.f, 50.f));
  CHECK_F3
    (f33_mul(dst, a, -1.f),
     f33(c, -1.f, -2.f, -3.f, -4.f, -5.f, -6.f, -7.f, -8.f, -9.f));
  CHECK_F33(dst, f33(c, -1.f, -2.f, -3.f, -4.f, -5.f, -6.f, -7.f, -8.f, -9.f));

  f33(b, 2.f, 9.f, 8.f, 1.f, -2.f, 2.f, 1.f, -8.f, -4.f);
  CHECK_F33
    (f33_mulf33(dst, a, b),
     f33(c, 94.f, 113.f, 132.f, 7.f, 8.f, 9.f, -59.f, -70.f,-81.f));
	CHECK_F33(dst, f33(c, 94.f, 113.f, 132.f, 7.f, 8.f, 9.f, -59.f, -70.f,-81.f));
  CHECK_F33
    (f33_mulf33(a, a, b),
     f33(c, 94.f, 113.f, 132.f, 7.f, 8.f, 9.f, -59.f, -70.f, -81.f));
	CHECK_F33(a, f33(c, 94.f, 113.f, 132.f, 7.f, 8.f, 9.f, -59.f, -70.f, -81.f));
  f33(a, 1.f, 2.f, 3.f, 4.f, 5.f, 6.f, 7.f, 8.f, 9.f);
  CHECK_F33
    (f33_mulf33(b, a, b),
     f33(c, 94.f, 113.f, 132.f, 7.f, 8.f, 9.f, -59.f, -70.f, -81.f));
	CHECK_F33(b, f33(c, 94.f, 113.f, 132.f, 7.f, 8.f, 9.f, -59.f, -70.f, -81.f));
  f33(b, 2.f, 9.f, 8.f, 1.f, -2.f, 2.f, 1.f, -8.f, -4.f);

  f33(a, 0.f, 1.f, 2.f, 3.f, 4.f, 5.f, 6.f, 7.f, 8.f);
  f33(b, 1.f, 2.f, 3.f, 4.f, 5.f, 6.f, 7.f, 8.f, 9.f);
  CHECK_F33
    (f33_add(dst, a, b),
     f33(c, 1.f, 3.f, 5.f, 7.f, 9.f, 11.f, 13.f, 15.f, 17.f));
  CHECK_F33(dst, f33(c, 1.f, 3.f, 5.f, 7.f, 9.f, 11.f, 13.f, 15.f, 17.f));
  CHECK_F33
    (f33_sub(dst, dst, b),
     f33(c, 0.f, 1.f, 2.f, 3.f, 4.f, 5.f, 6.f, 7.f, 8.f));
	CHECK_F33(dst, f33(c, 0.f, 1.f, 2.f, 3.f, 4.f, 5.f, 6.f, 7.f, 8.f));
  CHECK_F33
    (f33_minus(a, b),
     f33(c,  -1.f, -2.f, -3.f, -4.f, -5.f, -6.f, -7.f, -8.f, -9.f));
  CHECK_F33(a, f33(c,  -1.f, -2.f, -3.f, -4.f, -5.f, -6.f, -7.f, -8.f, -9.f));

  f33_set(a, b);
  CHECK(f33_eq(a, b), 1);
  f33_add(a, a, f33_splat(c, FLT_EPSILON));
  CHECK(f33_eq(a, b), 0);
  CHECK(f33_eq_eps(a, b, FLT_EPSILON), 1);
  CHECK(f33_eq_eps(a, b, FLT_EPSILON * 0.9f), 0);

  f33(a, 1.f, 2.f, 3.f, 4.f, 5.f, 6.f, 3.f, -4.f, 9.f);
  CHECK(f33_det(a), -60.f);
  CHECK(f33_inverse(b, a), -60.f);
  f33_mulf33(dst, a, b);
  CHECK(f33_eq_eps(dst, f33(c, 1.0f, 0.f, 0.f, 0.f, 1.f, 0.f, 0.f, 0.f, 1.f),
    1.e-6f), 1);
  CHECK(f33_invtrans(a, a), -60.f);
  CHECK_F33
    (a, f33(c, b[0], b[3], b[6], b[1], b[4], b[7], b[2], b[5], b[8]));

  f3(c, 0.666667f, 0.333333f, 0.666667f);
  f33_rotation_axis_angle(a, c, 0.349066f);
  CHECK(f33_eq_eps(a, f33(c,
    0.966496f, 0.241415f, -0.0872034f,
    -0.214612f, 0.946393f, 0.241415f,
    0.14081f, -0.214612f, 0.966496f), 1.e-6f), 1);

  return 0;
}
