/* Copyright (C) |Meso|Star> 2015-2016 (contact@meso-star.com)
 *
 * This software is a computer program whose purpose is to describe a
 * virtual 3D environment that can be ray-traced and sampled both robustly
 * and efficiently.
 *
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software. You can use,
 * modify and/or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty and the software's author, the holder of the
 * economic rights, and the successive licensors have only limited
 * liability.
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading, using, modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean that it is complicated to manipulate, and that also
 * therefore means that it is reserved for developers and experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and, more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms. */

#ifndef S3D_SCENE_C_H
#define S3D_SCENE_C_H

#include "s3d_backend.h"

#include <rsys/dynamic_array.h>
#include <rsys/hash_table.h>
#include <rsys/list.h>
#include <rsys/ref_count.h>

/*
 * The geometry pointers must be initialized to NULL in order to define
 * which pointers are valid or not
 */
static FINLINE void
geom_ptr_init__(struct mem_allocator* alloc, struct geometry** geom)
{
  (void)alloc; *geom = NULL;
}

/* Generate the darray_geom dynamic array */
#define DARRAY_NAME geom
#define DARRAY_DATA struct geometry*
#define DARRAY_FUNCTOR_INIT geom_ptr_init__
#include <rsys/dynamic_array.h>

/* Generate the htable_mesh hash table */
#define HTABLE_NAME geom
#define HTABLE_DATA struct geometry*
#define HTABLE_KEY struct s3d_shape*
#include <rsys/hash_table.h>

/* Generate the darray_fltui dynamic array */
struct fltui { float flt; unsigned ui; };
#define DARRAY_NAME fltui
#define DARRAY_DATA struct fltui
#include <rsys/dynamic_array.h>

/* Generate the darray_geom_nprims array */
struct nprims_cdf { unsigned nprims, irtc; };
#define DARRAY_NAME nprims_cdf
#define DARRAY_DATA struct nprims_cdf
#include <rsys/dynamic_array.h>

struct s3d_scene {
  struct list_node shapes; /* List of attached shapes */
  struct htable_geom cached_geoms; /* Cached shape geometries */
  struct darray_geom embree2geoms; /* Shape geometries index by embree id */
  struct darray_fltui cdf; /* Unormalized CDF */
  struct darray_nprims_cdf nprims_cdf;

  float lower[3], upper[3]; /* AABB of the scene */

  size_t instances_count; /* # instances in the scene */

  RTCScene rtc_scn; /* Embree scene */
  char is_rtc_scn_outdated; /* Must the embree scene rebuild */

  int session_mask; /* Combination of enum s3d_session_flag */

  struct s3d_device* dev;
  ref_T ref;
};

static FINLINE struct geometry*
scene_get_mesh(struct s3d_scene* scn, const unsigned igeom)
{
  struct geometry* geom;
  ASSERT(scn && igeom != RTC_INVALID_GEOMETRY_ID);
  ASSERT(igeom < darray_geom_size_get(&scn->embree2geoms));
  geom = darray_geom_data_get(&scn->embree2geoms)[igeom];
  ASSERT(geom);
  return geom;
}

#endif /* S3D_SCENE_C_H */

